

from raptor_functions.supervised.feature_extraction import  *
# from raptor_functions.unsupervised.train import *
# from raptor_functions.supervised.preprocess import  *
import optuna
import mlflow
import os

# try:
#     from sklearn.utils import safe_indexing
# except ImportError:
#     from sklearn.utils import _safe_indexing

from pathlib import Path
from mlflow.models.signature import infer_signature
from pycaret.clustering import *
from .ss_clustering import *
# from raptor_functions.semi_supervised.ss_clustering import *



# Importing the Packages:
import optuna
import mlflow
import os

# try:
#     from sklearn.utils import safe_indexing
# except ImportError:
#     from sklearn.utils import _safe_indexing

from pathlib import Path
from mlflow.models.signature import infer_signature
from explainerdashboard import ExplainerDashboard, ClassifierExplainer


















SENSORS_FEATURES = [
    "sensor_1",
    "sensor_2",
    "sensor_3",
    "sensor_4",
    "sensor_5",
    "sensor_6",
    "sensor_7",
    "sensor_8",
    "sensor_9",
    "sensor_10",
    "sensor_11",
    "sensor_12",
    "sensor_13",
    "sensor_14",
    "sensor_15",
    "sensor_16",
    "sensor_17",
    "sensor_18",
    "sensor_19",
    "sensor_20",
    "sensor_21",
    "sensor_22",
    "sensor_23",
    "sensor_24",
]

STAGES = ["baseline", "absorb", "pause", "desorb", "flush"]
TARGET_COL = "result"
REMOTE_TRACKING_URI = "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


# 'http://ec2-3-10-210-150.eu-west-2.compute.amazonaws.com:5000/'



def get_plots(model):

    model_attributes = dir(model)

    plot_dir = os.path.join(os.getcwd(), "plots")
    Path(plot_dir).mkdir(parents=True, exist_ok=True)

    try:
        plot_model(model, 'elbow', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'cluster', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'tsne', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'silhoutte', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'distance', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'distribution', save=plot_dir)
    except:
        pass

    
    
    
    
    

    
    mlflow.log_artifact(plot_dir)





def objective(trial, df, study_name, model_mode):



    mlflow.set_experiment(study_name)

    mlflow.set_tracking_uri(REMOTE_TRACKING_URI)


    unique_id = 'exp_unique_id'
    label = 'result'
    num_clusters = 2
    # model_name='kmeans'


    

    df_offset = df.filter(regex='offset')
    df_gradient = df.filter(regex='gradient')

    if len(df_offset.columns) > 0:
        offset = True
    else:
        offset = False

    if len(df_gradient.columns) > 0:
        gradient = True
    else:
        gradient = False

    mlflow.start_run()


    X = df.drop(TARGET_COL, axis=1)

    features = X
    relevant_features = list(X.columns)
    df_cluster, centroids = get_cluster_labels(df)

    model = SSC(centroids)

    # # remove unwanted characters from the column names
    # df_copy.columns = df_copy.columns.str.replace('["]', "")

    # features = df_copy.drop(TARGET_COL, axis=1)

    # df_copy["result"] = df_copy["result"].replace({"Control": 0, "Covid": 1})





    model_name = "ssc"
    

    # get_plots(model)

    # metrics = get_metrics()
    # print(metrics)
    # mlflow.log_metrics(metrics)


    # try:
        # signature = infer_signature(features, predict_model(model, data=features))
    signature = infer_signature(features, np.array(ss_predictions(X, centroids)))

    # mlflow.pyfunc.log_model(model)
    mlflow.pyfunc.log_model(artifact_path=model_name,python_model=model, signature=signature)


    # mlflow.pyfunc.log_model(artifact_path="model",python_model=ETS_Exogen, conda_env=conda_env)


        # mlflow.sklearn.log_model(model, artifact_path=model_name, signature=signature)
    # mlflow.pyfunc.log_model(model, artifact_path=model_name, signature=signature)


    # except:
    #     pass


    # model_params = model.get_params()



        
    

    




    # mlflow.log_metrics(metrics)

    try:
        # mlflow.set_tag('features', features)
        mlflow.set_tag("relevant_features", relevant_features)

    except:
        pass

    mlflow.set_tag("model_name", model_name)
    mlflow.set_tag("centroids", centroids)

    mlflow.set_tag("offset", offset)
    mlflow.set_tag("gradient", gradient)
    # mlflow.set_tag("model_params", model_params)


    # mlflow.set_tag("model_mode", model_mode)

    mlflow.end_run()


def train_experiments(
    df, study_name="semi-supervised", direction="maximize", model_mode="random", n_trials=5
):
    """trains several models during different trials and logs them. Experiemnt can be tracked on "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


    Args:
        df (pandas dataframe): dataframe of cyclic sensor data to be used for training
        study_name (str, optional): optuna study name to use. Defaults to 'unsupervised'.
        direction (str, optional): direction of objective. Defaults to 'maximize'.
        model_mode (str, optional): _description_. Defaults to 'random'.
        n_trials (int, optional): number of times to run experiments. Defaults to 5.
    """

    study = optuna.create_study(study_name=study_name, direction=direction)
    study.optimize(
        lambda trial: objective(trial, df, study_name, model_mode), n_trials=n_trials
    )

    print("Click on this link to track experiments: ", REMOTE_TRACKING_URI)






























# from raptor_functions.supervised.prediction import get_production_model, make_prediction_with_pm
from raptor_functions.supervised.datasets import get_data
import pandas as pd
from raptor_functions.supervised.prediction import make_ensemble_prediction, make_prediction_with_pm










unique_id = 'exp_unique_id'
label = 'result'





def label(row, preds):
    for i, pred in enumerate(preds):
        if i == row['exp_unique_id']:
            return pred




def get_pseudo_labeled_data(X):

    preds = make_prediction_with_pm(X)
    X[label] = X.apply(lambda row: label(row, preds), axis=1)

    return X

def retrain(df_labeled, df_unlabeled):

    X_unlabeled = df_unlabeled.drop(label, axis=1)
    pseudo_labeled_data = get_pseudo_labeled_data(X_unlabeled)
    df_all = pd.concat([df_labeled, pseudo_labeled_data])

    return df_all





#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function
from cop_kmeans import cop_kmeans, l2_distance
import argparse


def read_data(datafile):
    data = []
    with open(datafile, 'r') as f:
        for line in f:
            line = line.strip()
            if line != '':
                d = [float(i) for i in line.split()]
                data.append(d)
    return data

def read_constraints(consfile):
    ml, cl = [], []
    with open(consfile, 'r') as f:
        for line in f:
            line = line.strip()
            if line != '':
                line = line.split()
                constraint = (int(line[0]), int(line[1]))
                c = int(line[2])
                if c == 1:
                    ml.append(constraint)
                if c == -1:
                    cl.append(constraint)
    return ml, cl

def run(datafile, consfile, k, n_rep, max_iter, tolerance):
    data = read_data(datafile)
    ml, cl = read_constraints(consfile)

    best_clusters = None
    best_score = None
    for _ in range(n_rep):
        clusters, centers = cop_kmeans(data, k, ml, cl,
                                       max_iter=max_iter,
                                       tol=tolerance)
        if clusters is not None and centers is not None:
            score = sum(l2_distance(data[j], centers[clusters[j]])
                        for j in range(len(data)))
            if best_score is None or score < best_score:
                best_score = score
                best_clusters = clusters

    return best_clusters

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run COP-Kmeans algorithm')
    parser.add_argument('dfile', help='data file')
    parser.add_argument('cfile', help='constraint file')
    parser.add_argument('k', type=int, help='number of clusters')
    parser.add_argument('--ofile', help='file to store the output', default=None)
    parser.add_argument('--n_rep', help='number of times to repeat the algorithm',
                        default=10, type=int)
    parser.add_argument('--m_iter', help='maximum number of iterations of the main loop',
                        default=300, type=int)
    parser.add_argument('--tol', help='tolerance for deciding on convergence',
                        default=1e-4, type=float)
    args = parser.parse_args()

    clusters = run(args.dfile, args.cfile, args.k,
                            args.n_rep, args.m_iter, args.tol)

    if args.ofile is not None and clusters is not None:
        with open(args.ofile, 'w') as f:
            for cluster in clusters:
                f.write('%d\n' %cluster)

    if not clusters:
        print('No solution was found!')
    else:
        print(' '.join(str(c) for c in clusters))



import joblib
from io import BytesIO
import boto3
import mlflow
from mlflow.tracking import MlflowClient
from tsfresh.feature_extraction import settings, extract_features
# from .feature_extraction import add_offset_gradient



REMOTE_TRACKING_URI = 'http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/'



def get_model_and_features(model_uri, tracking_uri=REMOTE_TRACKING_URI):
    
    client = MlflowClient(tracking_uri=REMOTE_TRACKING_URI)

    
    # Load model as a Sklearn Model.
    loaded_model = mlflow.pyfunc.load_model(model_uri)
    run_id = loaded_model.metadata.run_id
    relevant_features = client.get_run(run_id).data.tags['relevant_features'].replace("'", '')[1:-1].split(', ')
    offset = bool(client.get_run(run_id).data.tags['offset'])
    gradient = bool(client.get_run(run_id).data.tags['gradient'])

    return loaded_model, relevant_features, offset, gradient




def get_prediction_features(
    X, model_uri, tracking_uri=REMOTE_TRACKING_URI,  id="exp_unique_id", timesteps="timesteps"
):


    
    _, relevant_features, offset, gradient = get_model_and_features(model_uri)

    # X = add_offset_gradient(X, offset=offset, gradient=gradient)
    # df = X.join(y)

    fc_parameters = settings.from_columns(relevant_features)

    prediction_data = extract_features(
        X, kind_to_fc_parameters=fc_parameters, column_id=id, column_sort=timesteps
    )

    prediction_data.columns = prediction_data.columns.str.replace('["]', "")


    return prediction_data


def get_production_model(name, tracking_uri=REMOTE_TRACKING_URI):

    client = MlflowClient(tracking_uri=tracking_uri)

    production_model_uri = client.get_latest_versions(name, ["Production"])[0].source

    loaded_model, relevant_features, offset, gradient = get_model_and_features(production_model_uri, tracking_uri=REMOTE_TRACKING_URI)

    return loaded_model, relevant_features, offset, gradient, production_model_uri


def make_prediction(df, model_uri, tracking_uri=REMOTE_TRACKING_URI,  id="exp_unique_id", timesteps="timesteps"):

    

    loaded_model, _, _, _ = get_model_and_features(model_uri, tracking_uri=REMOTE_TRACKING_URI)

    prediction_data = get_prediction_features(df, model_uri, tracking_uri,  id, timesteps)

    prediction = loaded_model.predict(prediction_data)

    return prediction


def make_prediction_with_pm(df, model_name='Covid Classifier', tracking_uri=REMOTE_TRACKING_URI,  
                            id="exp_unique_id", timesteps="timesteps"):

    loaded_model, relevant_features, offset, gradient, production_model_uri = get_production_model(model_name, 
                                                                                tracking_uri=REMOTE_TRACKING_URI)

    prediction_data = get_prediction_features(df, production_model_uri, 
                                                tracking_uri,  id, timesteps)

    prediction = loaded_model.predict(prediction_data)

    return prediction



def get_ensemble_model(name, tracking_uri=REMOTE_TRACKING_URI):

    client = MlflowClient(tracking_uri=tracking_uri)


    ensemble_uris = {}

    for m in client.search_model_versions(f"name='{name}'"):


        model_uri = m.source
        model_name = model_uri.split('/')[-1]

        # loaded_model, relevant_features, offset, gradient = get_model_and_features(model_uri, tracking_uri=REMOTE_TRACKING_URI)

        ensemble_uris[model_name] = model_uri



    return ensemble_uris


def get_ensemble_prediction(predictions):
    pred_sum = sum(predictions.values())
    if pred_sum > (len(predictions) / 2):
        return 1
    return 0


def make_ensemble_prediction(df, model_name='Ensemble Model', tracking_uri=REMOTE_TRACKING_URI,  id="exp_unique_id", timesteps="timesteps"):

    ensemble_dict = get_ensemble_model(model_name, tracking_uri=REMOTE_TRACKING_URI)

    preds = {}

    for model_name, model_uri in ensemble_dict.items():
        
        prediction = make_prediction(df, model_uri)

        preds[model_name] = prediction

    final_pred = get_ensemble_prediction(preds)

    return final_pred




# -*- coding: utf-8 -*-
import random

def cop_kmeans(dataset, k, ml=[], cl=[],
               initialization='kmpp',
               max_iter=300, tol=1e-4):

    ml, cl = transitive_closure(ml, cl, len(dataset))
    ml_info = get_ml_info(ml, dataset)
    tol = tolerance(tol, dataset)

    centers = initialize_centers(dataset, k, initialization)

    for _ in range(max_iter):
        clusters_ = [-1] * len(dataset)
        for i, d in enumerate(dataset):
            indices, _ = closest_clusters(centers, d)
            counter = 0
            if clusters_[i] == -1:
                found_cluster = False
                while (not found_cluster) and counter < len(indices):
                    index = indices[counter]
                    if not violate_constraints(i, index, clusters_, ml, cl):
                        found_cluster = True
                        clusters_[i] = index
                        for j in ml[i]:
                            clusters_[j] = index
                    counter += 1

                if not found_cluster:
                    return None, None

        clusters_, centers_ = compute_centers(clusters_, dataset, k, ml_info)
        shift = sum(l2_distance(centers[i], centers_[i]) for i in range(k))
        if shift <= tol:
            break

        centers = centers_

    return clusters_, centers_

def l2_distance(point1, point2):
    return sum([(float(i)-float(j))**2 for (i, j) in zip(point1, point2)])

# taken from scikit-learn (https://goo.gl/1RYPP5)
def tolerance(tol, dataset):
    n = len(dataset)
    dim = len(dataset[0])
    averages = [sum(dataset[i][d] for i in range(n))/float(n) for d in range(dim)]
    variances = [sum((dataset[i][d]-averages[d])**2 for i in range(n))/float(n) for d in range(dim)]
    return tol * sum(variances) / dim

def closest_clusters(centers, datapoint):
    distances = [l2_distance(center, datapoint) for
                 center in centers]
    return sorted(range(len(distances)), key=lambda x: distances[x]), distances

def initialize_centers(dataset, k, method):
    if method == 'random':
        ids = list(range(len(dataset)))
        random.shuffle(ids)
        return [dataset[i] for i in ids[:k]]

    elif method == 'kmpp':
        chances = [1] * len(dataset)
        centers = []

        for _ in range(k):
            chances = [x/sum(chances) for x in chances]
            r = random.random()
            acc = 0.0
            for index, chance in enumerate(chances):
                if acc + chance >= r:
                    break
                acc += chance
            centers.append(dataset[index])

            for index, point in enumerate(dataset):
                cids, distances = closest_clusters(centers, point)
                chances[index] = distances[cids[0]]

        return centers

def violate_constraints(data_index, cluster_index, clusters, ml, cl):
    for i in ml[data_index]:
        if clusters[i] != -1 and clusters[i] != cluster_index:
            return True

    for i in cl[data_index]:
        if clusters[i] == cluster_index:
            return True

    return False

def compute_centers(clusters, dataset, k, ml_info):
    cluster_ids = set(clusters)
    k_new = len(cluster_ids)
    id_map = dict(zip(cluster_ids, range(k_new)))
    clusters = [id_map[x] for x in clusters]

    dim = len(dataset[0])
    centers = [[0.0] * dim for i in range(k)]

    counts = [0] * k_new
    for j, c in enumerate(clusters):
        for i in range(dim):
            centers[c][i] += dataset[j][i]
        counts[c] += 1

    for j in range(k_new):
        for i in range(dim):
            centers[j][i] = centers[j][i]/float(counts[j])

    if k_new < k:
        ml_groups, ml_scores, ml_centroids = ml_info
        current_scores = [sum(l2_distance(centers[clusters[i]], dataset[i])
                              for i in group)
                          for group in ml_groups]
        group_ids = sorted(range(len(ml_groups)),
                           key=lambda x: current_scores[x] - ml_scores[x],
                           reverse=True)

        for j in range(k-k_new):
            gid = group_ids[j]
            cid = k_new + j
            centers[cid] = ml_centroids[gid]
            for i in ml_groups[gid]:
                clusters[i] = cid

    return clusters, centers

def get_ml_info(ml, dataset):
    flags = [True] * len(dataset)
    groups = []
    for i in range(len(dataset)):
        if not flags[i]: continue
        group = list(ml[i] | {i})
        groups.append(group)
        for j in group:
            flags[j] = False

    dim = len(dataset[0])
    scores = [0.0] * len(groups)
    centroids = [[0.0] * dim for i in range(len(groups))]

    for j, group in enumerate(groups):
        for d in range(dim):
            for i in group:
                centroids[j][d] += dataset[i][d]
            centroids[j][d] /= float(len(group))

    scores = [sum(l2_distance(centroids[j], dataset[i])
                  for i in groups[j])
              for j in range(len(groups))]

    return groups, scores, centroids

def transitive_closure(ml, cl, n):
    ml_graph = dict()
    cl_graph = dict()
    for i in range(n):
        ml_graph[i] = set()
        cl_graph[i] = set()

    def add_both(d, i, j):
        d[i].add(j)
        d[j].add(i)

    for (i, j) in ml:
        add_both(ml_graph, i, j)

    def dfs(i, graph, visited, component):
        visited[i] = True
        for j in graph[i]:
            if not visited[j]:
                dfs(j, graph, visited, component)
        component.append(i)

    visited = [False] * n
    for i in range(n):
        if not visited[i]:
            component = []
            dfs(i, ml_graph, visited, component)
            for x1 in component:
                for x2 in component:
                    if x1 != x2:
                        ml_graph[x1].add(x2)
    for (i, j) in cl:
        add_both(cl_graph, i, j)
        for y in ml_graph[j]:
            add_both(cl_graph, i, y)
        for x in ml_graph[i]:
            add_both(cl_graph, x, j)
            for y in ml_graph[j]:
                add_both(cl_graph, x, y)

    for i in ml_graph:
        for j in ml_graph[i]:
            if j != i and j in cl_graph[i]:
                raise Exception('inconsistent constraints between %d and %d' %(i, j))

    return ml_graph, cl_graph





from itertools import combinations, product
import numpy as np
import pandas as pd
from .copkmeans.cop_kmeans import cop_kmeans
import random
import mlflow.pyfunc
from scipy.spatial import distance_matrix
from scipy.spatial.distance import cdist


def make_random_choice(lst, links):
    new_list = []
    for x in range(links):
        choice = random.choice(lst)
        new_list.append(choice)
        if choice in new_list:
            lst.remove(choice)
    return new_list


def get_must_link(control_idx, covid_idx, links):
    control_comb = list(combinations(control_idx, 2))
    covid_comb = list(combinations(covid_idx, 2))
    
    must_link_control = make_random_choice(control_comb, links//2)
    must_link_covid = make_random_choice(covid_comb, links//2)

    return must_link_control + must_link_covid

  
def get_cannot_link(control_idx, covid_idx, links):
    covid_control_comb = list(product(covid_idx, control_idx))
    must_link_covid = make_random_choice(covid_control_comb, links)
    return must_link_covid



def get_constraints(df, links=4):

    control_idx = list(df[df['result'] == 'Control'].index)
    covid_idx = list(df[df['result'] == 'Covid'].index)

    must_link = get_must_link(control_idx, covid_idx, links)
    cannot_link = get_cannot_link(control_idx, covid_idx, links)
    
    return must_link, cannot_link

def replace_sample(df, column, fraction, val_replace):

    df_copy = df.copy()
    idx = df_copy[column].sample(frac=fraction).index
    df_copy.loc[idx, column] = val_replace

    return df_copy

def map(combined_df, label):
    df = combined_df[combined_df['result'] == label]
    cluster_value_counts = df['cluster'].value_counts()
    
    print('df: ', df)
    print('df_cluster: ', df['cluster'])
    print('label: ', label)
    print('cluster_value_counts: ', cluster_value_counts)
    print('cluster_value_value_type: ', type(cluster_value_counts))
    cluster = cluster_value_counts.idxmax()
    cluster_len = cluster_value_counts[cluster]
    if cluster_len / len(df) > 0.5:  
        return {cluster: label}
    return {cluster: -1}

def get_mappings(combined_df):
    
    map_dict = {}
    covid_map = map(combined_df, 'Covid')
    control_map = map(combined_df, 'Control')

    map_dict.update(control_map)
    map_dict.update(covid_map)

    return map_dict


def get_cluster_labels(df, k=2, links=4):

    must_link, cannot_link = get_constraints(df, links)

    print('must_link: ', must_link)
    print('cannot_link: ', cannot_link)
    X = df.drop('result', axis=1)

    input_matrix = X.values
    print('input_matrix: ', input_matrix)

    clusters, centers = cop_kmeans(dataset=input_matrix, k=k, ml=must_link, cl=cannot_link)

    print('clusters: ', clusters)
    print('centers: ', centers)
    print('clusters len: ', len(centers))
    print('clusters len: ', len(clusters))
    

    combined_df = df.join(pd.DataFrame(clusters, columns=['cluster']))

    map_dict = get_mappings(combined_df)

    combined_df['cluster_label'] = combined_df['cluster'].map(map_dict)

    return combined_df, centers






def ss_prediction(prediction_data, centroids):

    prediction_data = np.array(prediction_data)

    # print('Type: ', type(prediction_data))

    # print('N Shape: ', prediction_data.shape)
    # print('N Dimensions: ', prediction_data.ndim)

    # if prediction_data.ndim == 1:
    
    if prediction_data.ndim == 0:

        dist_matrix = pd.DataFrame(cdist(np.array(prediction_data).reshape(-1,1), centroids))
        cluster = dist_matrix.idxmin(axis=1)
        return cluster
    
    if prediction_data.ndim == 1:
        dist_matrix = pd.DataFrame(cdist(np.array(prediction_data).reshape(1,-1), centroids))
        cluster = dist_matrix.idxmin(axis=1)
        return cluster

        

def ss_predictions(prediction_data, centroids, mode='single'):
    # print('type: ', type(prediction_data))

    if (isinstance(prediction_data, pd.Series)) and mode=='multi':
        preds = prediction_data.apply(ss_prediction, centroids=centroids)
        return preds

    if (isinstance(prediction_data, pd.DataFrame)):
        preds = []
        for i in prediction_data.values:
        # print()
            pred = ss_prediction(i, centroids)
            preds.append(pred)
        return preds

    pred = ss_prediction(prediction_data, centroids)
    return pred
    
    # if prediction_data.ndim == 0:
    #     pred = ss_prediction(prediction_data, centroids)
    #     return pred
    # if prediction_data.ndim == 1:
    #     pred = ss_prediction(prediction_data, centroids)
    #     return pred

    # print(prediction_data)
    # return preds

# Define the model class
class SSC(mlflow.pyfunc.PythonModel):

    def __init__(self, centroids):
        self.centroids = centroids

    def predict(self, context, model_input):

        return ss_predictions(model_input, self.centroids)






import joblib
from io import BytesIO
import boto3
import mlflow
from mlflow.tracking import MlflowClient
from tsfresh.feature_extraction import settings, extract_features

from itertools import combinations, product
import numpy as np
import pandas as pd
from .copkmeans.cop_kmeans import cop_kmeans
import random
import mlflow.pyfunc
from scipy.spatial import distance_matrix
from scipy.spatial.distance import cdist

from supervised.feature_extraction import add_offset_gradient




REMOTE_TRACKING_URI = 'http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/'



def get_model_and_features(model_uri, tracking_uri=REMOTE_TRACKING_URI):
    
    client = MlflowClient(tracking_uri=REMOTE_TRACKING_URI)
    
    # Load model as a Sklearn Model.
    loaded_model = mlflow.pyfunc.load_model(model_uri)
    run_id = loaded_model.metadata.run_id
    relevant_features = client.get_run(run_id).data.tags['relevant_features'].replace("'", '')[1:-1].split(', ')
    offset = bool(client.get_run(run_id).data.tags['offset'])
    gradient = bool(client.get_run(run_id).data.tags['gradient'])

    return loaded_model, relevant_features, offset, gradient




def get_prediction_features(
    X, model_uri, tracking_uri=REMOTE_TRACKING_URI,  id="exp_unique_id", timesteps="timesteps"
):


    
    _, relevant_features, offset, gradient = get_model_and_features(model_uri)

    X = add_offset_gradient(X, offset=offset, gradient=gradient)
    # df = X.join(y)

    fc_parameters = settings.from_columns(relevant_features)

    prediction_data = extract_features(
        X, kind_to_fc_parameters=fc_parameters, column_id=id, column_sort=timesteps
    )

    prediction_data.columns = prediction_data.columns.str.replace('["]', "")


    return prediction_data




def ss_prediction(prediction_data, centroids):

    prediction_data = np.array(prediction_data)

    # print('Type: ', type(prediction_data))

    # print('N Shape: ', prediction_data.shape)
    # print('N Dimensions: ', prediction_data.ndim)

    # if prediction_data.ndim == 1:
    
    if prediction_data.ndim == 0:

        dist_matrix = pd.DataFrame(cdist(np.array(prediction_data).reshape(-1,1), centroids))
        cluster = dist_matrix.idxmin(axis=1)
        return cluster
    
    if prediction_data.ndim == 1:
        dist_matrix = pd.DataFrame(cdist(np.array(prediction_data).reshape(1,-1), centroids))
        cluster = dist_matrix.idxmin(axis=1)
        return cluster

        

def ss_predictions(prediction_data, centroids, mode='single'):
    # print('type: ', type(prediction_data))

    if (isinstance(prediction_data, pd.Series)) and mode=='multi':
        preds = prediction_data.apply(ss_prediction, centroids=centroids)
        return preds

    if (isinstance(prediction_data, pd.DataFrame)):
        preds = []
        for i in prediction_data.values:
        # print()
            pred = ss_prediction(i, centroids)
            preds.append(pred)
        return preds

    pred = ss_prediction(prediction_data, centroids)
    return pred


def make_prediction(df, model_uri, tracking_uri=REMOTE_TRACKING_URI,  id="exp_unique_id", timesteps="timesteps"):

    

    loaded_model, _, _, _ = get_model_and_features(model_uri, tracking_uri=REMOTE_TRACKING_URI)

    prediction_data = get_prediction_features(df, model_uri, tracking_uri,  id, timesteps)

    prediction = loaded_model.predict(prediction_data)

    return prediction




from raptor_functions.supervised.feature_extraction import  *
# from raptor_functions.unsupervised.train import *
# from raptor_functions.supervised.preprocess import  *
import optuna
import mlflow
import os

# try:
#     from sklearn.utils import safe_indexing
# except ImportError:
#     from sklearn.utils import _safe_indexing

from pathlib import Path
from mlflow.models.signature import infer_signature
from pycaret.clustering import *

# df = get_data('handheld_data')

# def get_clusters(df, model_name='kmeans', num_clusters=2):
#     unique_id = 'exp_unique_id'
#     label = 'result'

#     y = df.groupby(unique_id).first()[label]
#     X = df.drop(label, axis=1)

#     X = get_all_features(X)
#     df_extracted = X.join(y)

#     # df_f = get_training_features(df)


#     cluster = setup(X, session_id = 7652)

#     model = create_model(model_name, num_clusters)

#     # plot_model(model, 'elbow')
#     plot_model(model, 'cluster')


#     results = assign_model(model)
#     return results, model





# Importing the Packages:
import optuna
import mlflow
import os

# try:
#     from sklearn.utils import safe_indexing
# except ImportError:
#     from sklearn.utils import _safe_indexing

from pathlib import Path
from mlflow.models.signature import infer_signature
from explainerdashboard import ExplainerDashboard, ClassifierExplainer

SENSORS_FEATURES = [
    "sensor_1",
    "sensor_2",
    "sensor_3",
    "sensor_4",
    "sensor_5",
    "sensor_6",
    "sensor_7",
    "sensor_8",
    "sensor_9",
    "sensor_10",
    "sensor_11",
    "sensor_12",
    "sensor_13",
    "sensor_14",
    "sensor_15",
    "sensor_16",
    "sensor_17",
    "sensor_18",
    "sensor_19",
    "sensor_20",
    "sensor_21",
    "sensor_22",
    "sensor_23",
    "sensor_24",
]

STAGES = ["baseline", "absorb", "pause", "desorb", "flush"]
TARGET_COL = "result"
REMOTE_TRACKING_URI = "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


# 'http://ec2-3-10-210-150.eu-west-2.compute.amazonaws.com:5000/'


def get_plots(model):

    model_attributes = dir(model)

    plot_dir = os.path.join(os.getcwd(), "plots")
    Path(plot_dir).mkdir(parents=True, exist_ok=True)

    # plot_model(model, plot="confusion_matrix", save=plot_dir)
    # cm = os.path.join(plot_dir, "Confusion Matrix.png")
    # mlflow.log_artifact(cm)
    # log_artifact(cm)

    try:
        plot_model(model, 'elbow', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'cluster', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'tsne', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'silhoutte', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'distance', save=plot_dir)
    except:
        pass

    try:
        plot_model(model, 'distribution', save=plot_dir)
    except:
        pass

    
    
    
    
    

    
    mlflow.log_artifact(plot_dir)





def objective(trial, X, study_name, model_mode):



    mlflow.set_experiment(study_name)

    mlflow.set_tracking_uri(REMOTE_TRACKING_URI)


    unique_id = 'exp_unique_id'
    label = 'result'
    num_clusters = 2
    model_name='kmeans'

    # y = df.groupby(unique_id).first()[label]
    # X = df.drop(label, axis=1)

    # X = get_all_features(X)
    # df_extracted = X.join(y)

    # df_f = get_training_features(df)






    # results = assign_model(model)

    # df_offset = df.filter(regex='offset')
    # df_gradient = df.filter(regex='gradient')

    # if len(df_offset.columns) > 0:
    #     offset = True
    # else:
    #     offset = False

    # if len(df_gradient.columns) > 0:
    #     gradient = True
    # else:
    #     gradient = False
    

    # df_copy = df.copy(deep=True)

    # relevant_features = list(df_copy.drop(TARGET_COL, axis=1).columns)

    mlflow.start_run()

    # # remove unwanted characters from the column names
    # df_copy.columns = df_copy.columns.str.replace('["]', "")

    # features = df_copy.drop(TARGET_COL, axis=1)

    # df_copy["result"] = df_copy["result"].replace({"Control": 0, "Covid": 1})

    features = X
    relevant_features = list(X.columns)


    cluster = setup(X)

    MODELS = dict(models()["Name"])



    model_id = trial.suggest_categorical("model", list(MODELS.keys()))
    model_name = MODELS[model_id]
    model = create_model(model_id, num_clusters)
    
    # model = create_model(model_name, num_clusters)

    results = assign_model(model)




    MODELS = dict(models()["Name"])

    get_plots(model)

    metrics = get_metrics()
    # print(metrics)
    # mlflow.log_metrics(metrics)


    try:
        signature = infer_signature(features, predict_model(model, data=features))
        mlflow.sklearn.log_model(model, artifact_path=model_name, signature=signature)

    except:
        pass


    model_params = model.get_params()



        
    

    




    # mlflow.log_metrics(metrics)

    try:
        # mlflow.set_tag('features', features)
        mlflow.set_tag("relevant_features", relevant_features)

    except:
        pass

    mlflow.set_tag("model_name", model_name)
    # mlflow.set_tag("offset", offset)
    # mlflow.set_tag("gradient", gradient)
    mlflow.set_tag("model_params", model_params)


    mlflow.set_tag("model_mode", model_mode)

    mlflow.end_run()


def train_experiments(
    df, study_name="unsupervised", direction="maximize", model_mode="random", n_trials=5
):
    """trains several models during different trials and logs them. Experiemnt can be tracked on "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


    Args:
        df (pandas dataframe): dataframe of cyclic sensor data to be used for training
        study_name (str, optional): optuna study name to use. Defaults to 'unsupervised'.
        direction (str, optional): direction of objective. Defaults to 'maximize'.
        model_mode (str, optional): _description_. Defaults to 'random'.
        n_trials (int, optional): number of times to run experiments. Defaults to 5.
    """

    study = optuna.create_study(study_name=study_name, direction=direction)
    study.optimize(
        lambda trial: objective(trial, df, study_name, model_mode), n_trials=n_trials
    )

    print("Click on this link to track experiments: ", REMOTE_TRACKING_URI)
































# def plot_new_sample(df, x):

#     X = df.iloc[:, 1:13]
#     y = df["label"].replace({"Control": 0, "Covid": 1})  # .value_counts()
#     x = x.values.reshape(1, -1)

#     scaler = StandardScaler()
#     scaler.fit(X.values)
#     X = scaler.transform(X.values)

#     means = np.array(
#         [
#             243.94969928,
#             256.31424094,
#             313.33404167,
#             311.02738043,
#             284.31009239,
#             307.20553442,
#             293.71868478,
#             298.39285688,
#             276.90539674,
#             303.74205978,
#             262.39199638,
#             228.63297464,
#         ]
#     )
#     vars = np.array(
#         [
#             32.38231873,
#             15.39840351,
#             0.30922943,
#             0.11581567,
#             14.18706441,
#             0.67305964,
#             2.09758948,
#             0.96071839,
#             7.81709395,
#             1.53088058,
#             4.84237249,
#             13.75799395,
#         ]
#     )

#     # means = scaler.mean_
#     # vars = scaler.var_

#     print("mean: ", means)
#     print("var: ", vars)

#     # X = scale_data(X.values, means=means,stds=vars **0.5)
#     x = scale_data(x, means=means, stds=vars ** 0.5)

#     print("scaled_new_data: ", x)

#     pca = PCA(n_components=2)
#     X = pca.fit_transform(X)

#     # # print('Scaled X: ', X)

#     print("before pca: ", x)
#     x = pca.transform(x)
#     print("after pca: ", x)

#     # x = scaler.transform(x)
#     # print(x)

#     # pipe = Pipeline([
#     #     ('scale', StandardScaler()),
#     #     ('reduce_dims', PCA(n_components=2))])

#     # pipe.fit(X.values)
#     # X = pipe.transform(X.values)

#     # print('pca_comp: ', x)

#     group = y
#     # group = y[:-10]
#     cdict = {0: "blue", 1: "red"}
#     class_label = {0: "Control", 1: "Covid"}

#     scatter_x = X[:, 0]
#     scatter_y = X[:, 1]

#     fig, ax = plt.subplots()
#     for g in np.unique(group):
#         ix = np.where(group == g)
#         ax.scatter(
#             scatter_x[ix], scatter_y[ix], c=cdict[g], label=class_label[g], alpha=0.5
#         )
#     ax.scatter(x[:, 0], x[:, 1], c="black", label="new sample", s=70)
#     ax.legend()
#     ax.set_title("New data cluster assignment")
#     ax.set_xlabel("PC1")
#     ax.set_ylabel("PC2")
#     # plt.show()
#     fig.savefig("./static/img/sample_cluster.jpeg")



# !pip install raptor_functions pandas_profiling
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from pandas_profiling import ProfileReport
import sweetviz as sv
# from raptor_functions.datasets import get_data















import matplotlib
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    # 'text.usetex': True,
    'pgf.rcfonts': False,
})

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
# 
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns
# 
from sklearn.manifold import TSNE, Isomap, MDS
from sklearn.decomposition import PCA, KernelPCA
import umap
import os


def plot_sensors(df, all = True, exps_numbers = [1,2,3], sensors_numbers = [1,2,3,4,5,6,7,8,9,10,11,12], overlap = False, figure_name = 'sensor_response', path = os.getcwd()):
    """
    Returns a figure with the visualisation of the sensor response, and saves it as a .pdf figure
    
    Plot: the time dependent profiles of the sensor response [mV] against the elapsed time [s] for a wide range of combination of experiments.

    Parameters
    ----------
    df: pandas dataframe
        Dataframe containing all the experiments and sensor mesaurements. See raptor_functions 'preprocessing' functions to generate this dataframe from the raw .txt files
        df = [exp_unique_id, exp_name, timesteps, sensor_1 ... sensor_n, Humidity (r.h.), measurement_stage, date_exp,time_elapsed, datetime_exp_start,datetime_exp, filename,result]
    exps_numbers: list of integers
        Selection list of the experiments to be plotted. The 'exp_numbers' are given by the 'exp_unique_id' of the pandas dataframe 'df'.
    sensor_numbers: list of integers
        Selection list of the sensors to be plotted. The 'sensors_numbers' are given by the columns of the pandas dataframe 'df' - usually containing 24 sensor columns.
    overlap: boolean
        If overlap = True -> the different experiments (for one single sensor) are plotted in the same axis. Better for comparison of the response between different experiments.
        If overlap = False -> different experiments are plotted in different columns. Better to look at individual responses.
        Default: False
    figure_name: str
        Name to give the figure to be saved as .pdf
    path: str
        Absolute path where the figure will be saved. If nothing provided, saved in the working directory.
    
    Returns
    -------
    figure
        Outputs a matplolib object figure.
    file
        Saves a .pdf in the path with the figure

    """
    # -----------------------------------------------------------
    # 
    n_total_sensors = len(list(filter(lambda k: 'sensor' in k, df.columns.tolist())))
    n_total_exps = df['exp_unique_id'].nunique()
    # 
    if all == True:
        sensors_numbers = list(np.linspace(1, n_total_sensors, n_total_sensors, dtype= int))
        exps_numbers = list(np.linspace(0, n_total_exps-1, n_total_exps, dtype=int))  
    # 
    # Pre-processing
    experiments = []
    n_exp = len(exps_numbers)
    for i in range(len(exps_numbers)):
        df_exp = df.loc[df['exp_unique_id'] == exps_numbers[i]]
        experiments.append(df_exp)
    # 
    sensors = []
    n_subplots_y = len(sensors_numbers)
    for i in range(len(sensors_numbers)):
        sensor_number = sensors_numbers[i]
        sensor = 'sensor_' + str(sensor_number)
        sensors.append(sensor)
    # 
    legend = True
    # 
    if overlap == False and len(exps_numbers) == 1:
        overlap = True
    # 
    if len(exps_numbers) == 1 and len(sensors_numbers)==1:
        # 
        fig, ax = plt.subplots(figsize=(6, 2), tight_layout=True)
        ax.plot(df_exp['time_elapsed'], df_exp[sensor])
        ax.set_xlabel(r'{Time [s]}')
        ax.set_ylabel(r'{Response [mV]}')
        ax.set_title('Experiment ' + str(exps_numbers[0]) + ': ' + str(sensors[0]))
        # 

    elif overlap == True:
        alpha_val = 1
        linewidth_val = 1
        if n_exp >3:
            print("Warning: adding line tansparency (alpha = 0.7) and thinner lines (0.5) for enhaced visualisation.")
            alpha_val = 0.7
            linewidth_val = 0.5
        # 
        fig, axs = plt.subplots(n_subplots_y,1,figsize=(6, 1.5*n_subplots_y),sharex=True,tight_layout=True)
        for i in range(n_subplots_y):
            try:
                if i == 0:
                    axs[i].set_title('Experiments ' + str(exps_numbers) + '\n' + str(sensors[i]))
                else:
                    axs[i].set_title(str(sensors[i]))
                # 
                for j in range(n_exp):
                    axs[i].plot(experiments[j]['time_elapsed'], experiments[j][sensors[i]],label = str(exps_numbers[j]), alpha=alpha_val, linewidth = linewidth_val)
                    handles, labels = axs[i].get_legend_handles_labels()
                axs[i].set_ylabel(r'{Response [mV]}')
                if i == n_subplots_y -1:
                    axs[i].set_xlabel(r'{Time [s]}')
                    fig.legend(handles, labels, bbox_to_anchor=(1,0.5), loc="center left", borderaxespad=0)
            except: # Handling error when axis is 1D: axs[i]. -> axs.
                if i == 0:
                    axs.set_title('Experiments ' + str(exps_numbers) + '\n' + str(sensors[i]))
                else:
                    axs.set_title(str(sensors[i]))
                # 
                for j in range(n_exp):
                    axs.plot(experiments[j]['time_elapsed'], experiments[j][sensors[i]],label = str(exps_numbers[j]), alpha=alpha_val, linewidth = linewidth_val)
                    handles, labels = axs.get_legend_handles_labels()
                axs.set_ylabel(r'{Response [mV]}')
                if i == n_subplots_y -1:
                    axs.set_xlabel(r'{Time [s]}')
                    fig.legend(handles, labels, bbox_to_anchor=(1,0.5), loc="center left", borderaxespad=0)
                
        # 
    else:
        n_subplots_x = n_exp
        fig, axs = plt.subplots(n_subplots_y, n_subplots_x,figsize=(6*n_subplots_x, 1.5*n_subplots_y),sharex=True,tight_layout=True)
        for j in range(n_subplots_x):
            for i in range(n_subplots_y):
                try:
                    if i == 0:
                        axs[i,j].set_title('Experiment ' + str(exps_numbers[j]) + '\n' + str(sensors[i]))
                    else:
                        axs[i,j].set_title(str(sensors[i]))
                    # 
                    axs[i,j].plot(experiments[j]['time_elapsed'], experiments[j][sensors[i]])
                    axs[i,j].set_ylabel(r'{Response [mV]}')
                    if i == n_subplots_y -1:
                        axs[i,j].set_xlabel(r'{Time [s]}')
                except: # Above fails when n_sensors = 1 -> as the axs[i,j] become a 2d Array
                    if i == 0:
                        axs[j].set_title('Experiment ' + str(exps_numbers[j]) + '\n' + str(sensors[i]))
                    else:
                        axs[j].set_title(str(sensors[i]))
                    # 
                    axs[j].plot(experiments[j]['time_elapsed'], experiments[j][sensors[i]])
                    axs[j].set_ylabel(r'{Response [mV]}')
                    if i == n_subplots_y -1:
                        axs[j].set_xlabel(r'{Time [s]}')
        # 
    # -----------------------------------------------------------
    figure_name_full = figure_name + '.pdf'
    if path != os.getcwd():
        plt.savefig(path + "/" + figure_name_full, bbox_inches='tight')
        print('Plot saved sucesfully!')
    else: 
        plt.savefig(figure_name_full, bbox_inches='tight')
        print('Plot saved sucesfully!')

# 
# 
#


def PCA_plot(X,y,scaler = True, plt_3D = True, figure_name = 'pca_plot', path = os.getcwd()):
    # 
    if plt_3D == True:
        n_components = 3
    else:
        n_components = 2
    pca = PCA(n_components=n_components)
    if scaler == True:
        pipe = Pipeline([('scaler', StandardScaler()), ('pca', pca)])
        principalComponents = pipe.fit_transform(X)
    else: 
        principalComponents = pca.fit_transform(X)
    # 
    principalDf = pd.DataFrame(data = principalComponents)
    # 
    colors = {'Control':'green', 'Covid':'red'}
    # 
    # 
    if n_components == 2:
        fig, ax = plt.subplots(figsize=(6, 6), tight_layout=True)
        ax = sns.scatterplot(x=principalDf.iloc[:,0], y=principalDf.iloc[:,1], hue=y, alpha=0.9, edgecolor='black')
    elif n_components == 3:
        # 
        fig = plt.figure(figsize=(6,6))    
        ax = Axes3D(fig, auto_add_to_figure=False)
        fig.add_axes(ax)    
        cmap = ListedColormap(sns.color_palette("husl", 256).as_hex())
        sc = ax.scatter(principalDf.iloc[:,0],principalDf.iloc[:,1],principalDf.iloc[:,2], c=y.map(colors), cmap=cmap, alpha=0.9, edgecolors='black')
        ax.set_zlabel('Axis 3')
    else:
        print('Error: please only select 2 or 3 axis for the Principal Component Analysis.')
    ax.set_xlabel('Axis 1')
    ax.set_ylabel('Axis 2')
    ax.set_title('Principal Component Analysis (PCA)')
    # -----------------------------------------------------------
    # TODO: create new directory if path specified doesn't exists
    figure_name_full = figure_name + '.pdf'
    if path != os.getcwd():
        plt.savefig(path + "/" + figure_name_full, bbox_inches='tight')
        print('Plot saved sucesfully!')
    else: 
        plt.savefig(figure_name_full, bbox_inches='tight')
        print('Plot saved sucesfully!')


# 
# 
# 

def dimension_reduction_plot(X,y,scaler = True, plt_3D = True, single_plot = True, figure_name = 'dr_plot', path = os.getcwd()):
    # 
    if plt_3D == True:
        n_components = 3
    else:
        n_components = 2
    # Methods
    import umap
    pca = PCA(n_components=n_components)
    mds = MDS(n_components=n_components)
    umap = umap.UMAP(n_components=n_components)
    tsne = TSNE(n_components=n_components)
    isomap = Isomap(n_components=n_components)
    kernelpca = KernelPCA(n_components=n_components)
    # 
    if scaler == True:
        scaled_X = StandardScaler().fit_transform(X)
        # 
        pc_pca = pca.fit_transform(scaled_X)
        pc_mds = mds.fit_transform(scaled_X)
        pc_umap = umap.fit_transform(scaled_X)
        pc_tsne = tsne.fit_transform(scaled_X)
        pc_isomap = isomap.fit_transform(scaled_X)
        pc_kernelpca = kernelpca.fit_transform(scaled_X)
    else: 
        pc_pca = pca.fit_transform(X)
        pc_mds = mds.fit_transform(X)
        pc_umap = umap.fit_transform(X)
        pc_tsne = tsne.fit_transform(X)
        pc_isomap = isomap.fit_transform(X)
        pc_kernelpca = kernelpca.fit_transform(X)
    # 
    df_pc_pca = pd.DataFrame(data = pc_pca)
    df_pc_mds = pd.DataFrame(data = pc_mds)
    df_pc_umap = pd.DataFrame(data = pc_umap)
    df_pc_tsne = pd.DataFrame(data = pc_tsne)
    df_pc_isomap = pd.DataFrame(data = pc_isomap)
    df_pc_kernelpca = pd.DataFrame(data = pc_kernelpca)
    # 
    df_pc_list = [df_pc_pca, df_pc_mds, df_pc_umap, df_pc_tsne, df_pc_isomap, df_pc_kernelpca]
    df_pc_name_list = ['PCA', 'MDS', 'UMAP', 't-SNE', 'isoMap', 'Kernel PCA']
    n_subplots = len(df_pc_list)
    # 
    colors = {'Control':'green', 'Covid':'red'}
    # 
    # 
    if n_components == 2:
        if single_plot == True:
            fig, ax = plt.subplots(n_subplots,1,figsize=(6, 5*n_subplots),tight_layout=True)
            for i in range(n_subplots):
                df_pc_temp = df_pc_list[i]
                df_pc_name_list_temp = df_pc_name_list[i]
                sns.scatterplot(x=df_pc_temp.iloc[:,0], y=df_pc_temp.iloc[:,1], hue=y, alpha=0.9, edgecolor='black', ax=ax[i])
                ax[i].set_title(df_pc_name_list_temp)
            plt.show()
        else:
            for i in range(len(df_pc_list)):
                df_pc_temp = df_pc_list[i]
                df_pc_name_list_temp = df_pc_name_list[i]
                # 
                fig, ax = plt.subplots(figsize=(6, 6), tight_layout=True)
                ax = sns.scatterplot(x=df_pc_temp.iloc[:,0], y=df_pc_temp.iloc[:,1], hue=y, alpha=0.9, edgecolor='black')
                ax.set_xlabel('Axis 1')
                ax.set_ylabel('Axis 2')
                ax.set_title('Dimensional Reduction: ' + df_pc_name_list_temp)
                # -----------------------------------------------------------
                # TODO: create new directory if path specified doesn't exists
                figure_name_full = figure_name + '_' + str(df_pc_name_list_temp) + '.pdf'
                if path != os.getcwd():
                    plt.savefig(path + "/" + figure_name_full, bbox_inches='tight')
                    print('Plot saved sucesfully!')
                else: 
                    plt.savefig(figure_name_full, bbox_inches='tight')
                    print('Plot saved sucesfully in current path!')
                plt.show()
    elif n_components == 3:
        if single_plot == True:
            fig, ax = plt.subplots(figsize=(6, 5*n_subplots),tight_layout=True)
            plt.axis('off')
            for i in range(n_subplots):
                df_pc_temp = df_pc_list[i]
                df_pc_name_list_temp = df_pc_name_list[i]
                # 
                ax = fig.add_subplot(n_subplots, 1, i+1,projection='3d')
                cmap = ListedColormap(sns.color_palette("husl", 256).as_hex())
                ax.scatter(df_pc_temp.iloc[:,0],df_pc_temp.iloc[:,1],df_pc_temp.iloc[:,2], c=y.map(colors), cmap=cmap, alpha=0.9, edgecolors='black')
                ax.set_zlabel('Axis 3')
                ax.set_xlabel('Axis 1')
                ax.set_ylabel('Axis 2')
                ax.set_title(df_pc_name_list_temp)
            # -----------------------------------------------------------
            # TODO: create new directory if path specified doesn't exists
            figure_name_full = figure_name + '.pdf'
            if path != os.getcwd():
                plt.savefig(path + "/" + figure_name_full, bbox_inches='tight')
                print('Plot saved sucesfully!')
            else: 
                plt.savefig(figure_name_full, bbox_inches='tight')
                print('Plot saved sucesfully in current path!')
            plt.show()

        else:   
            for i in range(len(df_pc_list)):
                df_pc_temp = df_pc_list[i]
                df_pc_name_list_temp = df_pc_name_list[i]
                # 
                fig = plt.figure(figsize=(6,6))    
                ax = Axes3D(fig, auto_add_to_figure=False)
                fig.add_axes(ax)    
                cmap = ListedColormap(sns.color_palette("husl", 256).as_hex())
                sc = ax.scatter(df_pc_temp.iloc[:,0],df_pc_temp.iloc[:,1],df_pc_temp.iloc[:,2], c=y.map(colors), cmap=cmap, alpha=0.9, edgecolors='black')
                ax.set_zlabel('Axis 3')
                ax.set_xlabel('Axis 1')
                ax.set_ylabel('Axis 2')
                ax.set_title('Dimensional Reduction: ' + df_pc_name_list_temp)
                # -----------------------------------------------------------
                # TODO: create new directory if path specified doesn't exists
                figure_name_full = figure_name + '_' + str(df_pc_name_list_temp) + '.pdf'
                if path != os.getcwd():
                    plt.savefig(path + "/" + figure_name_full, bbox_inches='tight')
                    print('Plot saved sucesfully!')
                else: 
                    plt.savefig(figure_name_full, bbox_inches='tight')
                    print('Plot saved sucesfully in current path!')
                plt.show()
    else:
        print('Error: please only select 2 or 3 axis for the Principal Component Analysis.')



def eda_sv(df, compare=False):

    df['result'] = df['result'].map({'Control':0, 'Covid':1})



    if compare:
        covid = df[df['result']==1]
        control = df[df['result']==0]
        my_report = sv.compare([covid, "Covid"], [control, "Control"], 'result')
        my_report.show_notebook(w=None, h=None, scale=None, layout='widescreen', filepath=None)
    
    else:
        my_report = sv.analyze(df, target_feat='result', pairwise_analysis='off')
        my_report.show_notebook(w=None, h=None, scale=None, layout='widescreen', filepath=None)





def eda_pp(df):
    profile = ProfileReport(df, title="Pandas Profiling Report", explorative=True)
    # profile.to_file("pandas_profiling_report.html")
    # profile.to_widgets()
    profile.to_notebook_iframe()

# import ssl
# ssl._create_default_https_context = ssl._create_unverified_context
# #
# df = get_data()
# #
# from pandas_profiling import ProfileReport
# profile = ProfileReport(df, title="Pandas Profiling Report", explorative=True)
# profile.to_file("pandas_profiling_report.html")



# import sweetviz as sv
# my_report = sv.analyze(df)
# my_report.show_html() # Default arguments will generate to "SWEETVIZ_REPORT.html"



# from autoviz.AutoViz_Class import AutoViz_Class
# AV = AutoViz_Class()

# import nltk
# nltk.download('wordnet')

# _ = AV.AutoViz('df.csv')




# import dtale
# import dtale.app as dtale_app
# # dtale_app.USE_COLAB = True
# dtale_app.USE_NGROK = True
# dtale.show(df)